package fr.bhprod.sga.path;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import com.jme.scene.Node;

/**
 * 
 * @author jbeaufils
 * @version 1.0
 */

public class AStar {
	public static final int STRAIGHT_COST = 10;
	public static final int DIAG_COST = 14;
	
	public static final Logger LOGGER = Logger.getLogger(AStar.class);

	public static final Vector<int[]> calculateBasicAStar(int startx, int starty, int destx, int desty, int sizex, int sizey, short[][] walkable, Node rootNode) {
		LOGGER.debug("start : "+startx+"x"+starty);
		LOGGER.debug("end : "+destx+"x"+desty);
		
		Hashtable<String, PathTile> listeOuverte = new Hashtable<String, PathTile>();
		Hashtable<String, PathTile> listeFermee = new Hashtable<String, PathTile>();

		PathTile startTile = new PathTile(null, startx, starty, 0, calculateManhattanDistance(startx, starty, destx, desty));
		
		listeOuverte.put(startTile.getName(), startTile);

		// Start of treating the open list
		int count = 0;
		
		while (!listeOuverte.isEmpty() && count < 300) {
			int minF = Integer.MAX_VALUE;
			PathTile curTile = null;
			
			//LOGGER.debug("liste ouverte : "+listeOuverte);

			for (Iterator<Entry<String,PathTile>> it = listeOuverte.entrySet().iterator(); it.hasNext();) {
				PathTile tile = it.next().getValue();
				if (tile.getF() < minF) {
					minF = tile.getF();
					curTile = tile;
				}
			}
			
			//LOGGER.debug("curTile : "+curTile);

			listeOuverte.remove(curTile.getName());
			listeFermee.put(curTile.getName(), curTile);
			
			if (curTile.getX() == destx && curTile.getY() == desty) {
				Vector<int[]> results = new Vector<int[]>();
				results.add(0, new int[] {curTile.getX(), curTile.getY()});
				while (curTile.getParent() != null) {
					results.add(0, new int[] {curTile.getX(), curTile.getY()});
					curTile = curTile.getParent();
				}
				LOGGER.debug("Destination atteinte !\n");
				return results;
			} else {
				/*Sphere sphere = new Sphere("jhbjh"+curTile.getX()+curTile.getY(), new Vector3f(curTile.getX() * 2 + 1, 10, -(curTile.getY() * 2 + 1)), 20, 20, 0.5f);
				rootNode.attachChild(sphere);*/
			}
			
			PathTile[] voisins = findNeighbours(curTile, destx, desty, sizex, sizey, walkable);
			
			int newG = 0;
			PathTile oldTile = null;
	
			for (int i = 0; i < voisins.length; i++) {
				if (! listeFermee.containsKey(voisins[i].getName())) {
					if (listeOuverte.containsKey(voisins[i].getName())) {
						newG = curTile.getG();
						if (voisins[i].getX() == curTile.getX() + 1 ||
								voisins[i].getX() == curTile.getX() - 1 ||
								voisins[i].getY() == curTile.getY() + 1 ||
								voisins[i].getY() == curTile.getY() - 1) {
							newG += DIAG_COST;
						} else {
							newG += STRAIGHT_COST;
						}
						oldTile = listeOuverte.get(voisins[i].getName());
						if (newG < oldTile.getG()) {
							oldTile.setParent(curTile);
							oldTile.setG(newG);
							listeOuverte.put(oldTile.getName(), oldTile);
						}
					} else {
						listeOuverte.put(voisins[i].getName(), voisins[i]);
					}
				}
			}
			
			count++;
		}
		// End of treating the open list

		LOGGER.warn("Impossible d'atteindre la destination !");
		return null;
	}
	
	private static final PathTile[] findNeighbours(PathTile curTile, int xDest, int yDest, int sizeX, int sizeY, short[][] walkable) {
		Vector<PathTile> tiles = new Vector<PathTile>();
		int x = curTile.getX();
		int y = curTile.getY();
		int g = curTile.getG();
		
		if (x == 0 && y == 0) {
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
			if (walkable[x+1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y+1,  g + DIAG_COST, calculateManhattanDistance(x+1, y+1, xDest, yDest)));
			}
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
		} else if (x == 0 && y == sizeY -1) {
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
			if (walkable[x+1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y-1,  g + DIAG_COST, calculateManhattanDistance(x+1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
		} else if (x == sizeX -1 && y == 0) {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y+1,  g + DIAG_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
		} else if (x == sizeX -1 && y == sizeY -1) {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y-1,  g + DIAG_COST, calculateManhattanDistance(x-1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
		} else if (x == 0) {
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y+1,  g + DIAG_COST, calculateManhattanDistance(x+1, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
			if (walkable[x+1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y-1,  g + DIAG_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
		} else if (x == sizeX -1) {
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
			if (walkable[x-1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y+1,  g + DIAG_COST, calculateManhattanDistance(x-1, y+1, xDest, yDest)));
			}
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y-1,  g + DIAG_COST, calculateManhattanDistance(x-1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
		} else if (y == 0) {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y+1,  g + DIAG_COST, calculateManhattanDistance(x-1, y+1, xDest, yDest)));
			}
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y+1,  g + DIAG_COST, calculateManhattanDistance(x+1, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
		} else if (y == sizeY -1) {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y-1,  g + DIAG_COST, calculateManhattanDistance(x-1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
			if (walkable[x+1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y-1,  g + DIAG_COST, calculateManhattanDistance(x+1, y-1, xDest, yDest)));
			}
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
		} else {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y+1,  g + DIAG_COST, calculateManhattanDistance(x-1, y+1, xDest, yDest)));
			}
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y+1,  g + DIAG_COST, calculateManhattanDistance(x+1, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
			if (walkable[x+1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y-1,  g + DIAG_COST, calculateManhattanDistance(x+1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
			if (walkable[x-1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y-1,  g + DIAG_COST, calculateManhattanDistance(x-1, y-1, xDest, yDest)));
			}
		}
		
		PathTile[] result = new PathTile[tiles.size()];
		result = tiles.toArray(result);
		return (result);
	}
	
	private static final int calculateManhattanDistance(int xOri, int yOri, int xDest, int yDest) {
		return (((xDest>xOri?xDest-xOri:xOri-xDest) + (yDest>yOri?yDest-yOri:yOri-yDest)) * STRAIGHT_COST);
	}
}

